require "process_slicer"

-- From lua-users RangeIterator
-- range(start)             returns an iterator from 1 to a (step = 1)
-- range(start, stop)       returns an iterator from a to b (step = 1)
-- range(start, stop, step) returns an iterator from a to b, counting by step.
function range(i, to, inc)
  if i == nil then return end -- range(--[[ no args ]]) -> return "nothing" to fail the loop in the caller

  if not to then
    to = i
    i  = to == 0 and 0 or (to > 0 and 1 or -1)
  end

  -- we don't have to do the to == 0 check
  -- 0 -> 0 with any inc would never iterate
  inc = inc or (i < to and 1 or -1)

  -- step back (once) before we start
  i = i - inc

  return function () if i == to then return nil end i = i + inc return i, i end
end

-- Adapted from a RedBlob Tutorial
-- TODO Add Yellow noise (?)
function noise(mapsize, freq)
  local phase = math.random(0, 2*math.pi)
  local data = {}
  for x in range(mapsize) do
    data[x] = math.sin(2*math.pi * freq*x/mapsize + phase)
  end
  return data
end

function weighted_sum(mapsize, amplitudes, noises)
  --assert(#amplitudes > mapsize and #noises > mapsize)
  local output={}
  for i in range(mapsize) do
    output[i]=0.0
  end
  for k in range(#noises) do
    for x in range(mapsize) do
      output[x] = output[x] + (amplitudes[k] * noises[k][x])
    end
  end
  return output
end

-- generates all those noise tables, coroutine version
function random_ift(length, frequencies, amplitude, update_progress)
  local amplitudes = {}
  local noises = {}
  for f=1,frequencies do
    amplitudes[f]=amplitude(f)
    noises[f]=noise(length, f)

    update_progress(f / frequencies)

    coroutine.yield()
  end
  return weighted_sum(length, amplitudes, noises)
end

-- non-coroutine version
function random_ift_old(length, frequencies, amplitude, update_progress)
  local amplitudes = {}
  local noises = {}
  for f=1,frequencies do
    amplitudes[f]=amplitude(f)
    noises[f]=noise(length, f)
  end
  return weighted_sum(length, amplitudes, noises)
end
-- End adaptation

function random_ift_dialog(length, frequencies, amplitude)
  local dialog, process
  local vb = renoise.ViewBuilder()

  local function update_progress(progress)
    if (not dialog or not dialog.visible) then
      -- abort processing when the dialog was closed
      process:stop()
      return
    end
    
    -- else update the progress text
    if (progress == 1.0) then
      dialog:close()
    else
      vb.views.progress_text.text = string.format(
        "Processing frequencies (%d Hz)...", progress * frequencies)
    end
  end

  local function cancel_process()
    process:stop()
    dialog:close()
  end

  local DEFAULT_DIALOG_MARGIN = 
    renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
  
  local DEFAULT_CONTROL_SPACING = 
    renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
  
  local DEFAULT_DIALOG_BUTTON_HEIGHT = 
    renoise.ViewBuilder.DEFAULT_DIALOG_BUTTON_HEIGHT

  local dialog_content = vb:column {
    uniform = false,
    margin = DEFAULT_DIALOG_MARGIN,
    spacing = DEFAULT_CONTROL_SPACING,

    -- (add some content here)
    
    vb:text {
      id = "progress_text",
      text = "Getting ready to process some frequencies..."
    },

    vb:button {
      id = "cancel_button",
      text = "Cancel",
      height = DEFAULT_DIALOG_BUTTON_HEIGHT,
      width = 80,
      notifier = cancel_process
    }
  }

  process = ProcessSlicer(random_ift, length, frequencies, amplitude, update_progress)
  process:start()
  dialog = renoise.app():show_custom_dialog(
    "Processing Frequencies", dialog_content)

  return process
end

local exp_ampl_func = function(exp) return function(freq) return math.pow(freq, exp) end end
local loggy_ampl = function(freq) return math.log(freq) end
local expon_ampl = function(base) return function(freq) return math.pow(base, freq) end end
local e_ampl_func = function(exp) return function(freq) return math.pow(math.exp(freq), exp) end end

--TODO Look into using coroutines to generate the data as the GUI is a bit slow (maybe? prob not tho)

function prepare_data(buff)
  buff:prepare_sample_data_changes()
end

function finalize_data(buff)
  buff:finalize_sample_data_changes()
end

function send_data(buff, chan, frame, data)
  if chan == renoise.SampleBuffer.CHANNEL_LEFT_AND_RIGHT then
    buff:set_sample_data(renoise.SampleBuffer.CHANNEL_LEFT, frame, data)
    buff:set_sample_data(renoise.SampleBuffer.CHANNEL_RIGHT, frame, data)
  else
    buff:set_sample_data(chan, frame, data)
  end
end

-- generators function format:
-- buffer, sample_start, sample_end, length, channel, type, frequencies, periods, volume, anticlick, duty

function noise_gen(buff, ss, se, length, chan, type, freqs, _, volume)
  prepare_data(buff)
  local types = {exp_ampl_func(-2),
  exp_ampl_func(-1),
  exp_ampl_func(0),
  exp_ampl_func(1),
  exp_ampl_func(2),
  loggy_ampl,
  expon_ampl(2),
  e_ampl_func(-1),
  }
  --oprint("type:"..tostring(type))
  local ampl_func = types[type]

  if freqs >= 250.0 then
    -- pop open the dialog (interrupt the user)
    local data_process = random_ift_dialog(length, freqs, ampl_func)

    local function post_process()
      if data_process.return_value == nil then
        if data_process:running() then
          -- process still processing
          return
        else
          -- process was cancelled for some reason
          renoise.tool().app_idle_observable:remove_notifier(post_process)
          return
        end
      end

      local data = data_process.return_value

      print(data)

      -- Now how the FUCK do we wait until everything is done to continue processing

      --local data = random_ift(length, freqs, ampl_func) -- see if we can open a dialog for random_ift
      local top = 0
      for f=1,freqs do
        top = top + ampl_func(f)
      end
      for f in range(ss, se) do
        send_data(buff, chan, f, (data[f - ss + 1] / top) * volume)
      end
      finalize_data(buff)

      renoise.tool().app_idle_observable:remove_notifier(post_process)
    end

    renoise.tool().app_idle_observable:add_notifier(post_process)
  else
    -- DON'T open a dialog, and keep the experience smooth as butta
    local data = random_ift_old(length, freqs, ampl_func, function() end)

    local top = 0
    for f=1,freqs do
      top = top + ampl_func(f)
    end
    for f in range(ss, se) do
      send_data(buff, chan, f, (data[f - ss + 1] / top) * volume)
    end
    finalize_data(buff)
  end
end

function sine_gen(buff, ss, se, length, chan, _, _, periods, volume)
  prepare_data(buff)
  for f in range(ss, se) do
    send_data(buff, chan, f, math.sin(periods * 2*math.pi*((f-ss)/length)) * volume)
  end
  finalize_data(buff)
end

function square_gen(buff, ss, se, length, chan, _, _, periods, volume)
  prepare_data(buff)
  for f in range(ss, se) do
    send_data(buff, chan, f, ((math.sin(periods * 2*math.pi*((f-ss)/length)) < 0) and -1 or 1) * volume)
  end
  finalize_data(buff)
end

function pulse_gen(buff, ss, se, length, chan, _, _, periods, volume, _, duty)
  prepare_data(buff)
  -- duty is a number between 0 and 1, where 0.5 is square wave
  for f in range(ss, se) do
    -- Adapted saw code to properly do pulse generation
    local inp = _calc_x(f, ss, length, periods, 2, false) * math.pi * periods
    local samp = (-2/math.pi) * math.atan(math.cos(inp) /  math.sin(inp))
    local psamp = (samp < (duty * 2 - 1) and -1 or 1) * volume
    send_data(buff, chan, f, psamp)
  end
  finalize_data(buff)
end

function _calc_x(f, ss, length, periods, adj, ac)
  local x = ((f-ss)/length)
  if ac then
    x = x + (1/(adj*periods))
  end
  return x
end

function triangle_gen(buff, ss, se, length, chan, _, _, periods, volume, ac, _)
  prepare_data(buff)
  for f in range(ss, se) do
     -- Start from 0
    local inp = _calc_x(f, ss, length, periods, 4, ac) * math.pi * periods
    local saw = (-2/math.pi) * math.atan(math.cos(inp) /  math.sin(inp))
    local samp = (2*math.abs(saw)-1) * volume
    send_data(buff, chan, f, samp)
  end
  finalize_data(buff)
end

function sawtooth_gen(buff, ss, se, length, chan, _, _, periods, volume, ac, _)
  prepare_data(buff)
  for f in range(ss, se) do
    -- Inspired by formula from Wikipedia
    local inp = _calc_x(f, ss, length, periods, 2, ac) * math.pi * periods
    local samp = (-2*volume/math.pi) * math.atan(math.cos(inp) /  math.sin(inp))
    send_data(buff, chan, f, samp)
  end
  finalize_data(buff)
end

function process_generator(generator_num, gopt)
  rprint(gopt)
  print("Called gen #"..tostring(generator_num))
  local generators = {noise_gen, sine_gen, square_gen, triangle_gen, sawtooth_gen, pulse_gen}
  local inst_index = renoise.song().selected_instrument_index
  local samp_index = renoise.song().selected_sample_index

  local inst = renoise.song():instrument(inst_index)

  local sample
  if samp_index == 0 then
    sample = inst:insert_sample_at(#inst.samples+1)
  else
    sample = renoise.song():instrument(inst_index):sample(samp_index)
  end

  -- Always set the loop mode to FORWARD, not just when we create samples
  -- If the user wants to change it, they can easily do so.
  if sample.loop_mode == renoise.Sample.LOOP_MODE_OFF then
    sample.loop_mode = renoise.Sample.LOOP_MODE_FORWARD
  end
  local sample_buff = sample.sample_buffer

    -- Real anticlick
  sample.autofade = gopt.anticlick

  if not sample_buff.has_sample_data then
    local options = renoise.tool().preferences
    sample_buff:create_sample_data(
      options.sample_rate.value,
      options.bit_depth.value,
      options.num_channels.value,
      options.num_frames.value)
  end
  local sse = sample_buff.selection_range
  local chan
  if sample_buff.number_of_channels > 1 then
    chan = sample_buff.selected_channel
  else
    chan = renoise.SampleBuffer.CHANNEL_LEFT
  end
  local length = sse[2] - sse[1] + 1
  generators[generator_num](
    sample_buff,
    sse[1],
    sse[2],
    length,
    chan,
    gopt.noise,
    gopt.frequencies,
    gopt.period,
    gopt.volume/100.0,
    gopt.anticlick,
    gopt.duty
  )
end

function generate_init()
  math.randomseed(os.clock()) -- Initalize random generator for session
end
