--[[
 The goal of this project is to port MilkyTracker's Generators (Noise, Sine, Square, Triangle and Sawtooth)
 over to Renoise as closely as possible. That mainly includes

 Noise:
  Asking for either Pink, Brown or White noise

 Sine, Square, Triangle, and Sawtooth:
  Asking for Volume and Periods

  All:
    Generating given waveform for entire sample unless there is a selection,
    otherwise only generating it for that section
]]--

require "generate"

_AUTO_RELOAD_DEBUG = true

local presets = renoise.Document.create("MilkyWaves") {
  noise = 1, -- [1, 3]
  period = 1.0, -- (0.0, 2000.0]
  volume = 100.0, -- [-100.0, 100.0] maps to [-1, 1.0]
  frequencies = 30,
  anticlick = true,
  duty = 0.5, -- (0, 1) maps to (-1, 1)

  -- SampleBuffer generation shtuff (todo make configurable in *separate* dialog)
  sample_rate = 44100,
  bit_depth = 16,
  num_channels = 1,
  num_frames = 168,
}

renoise.tool().preferences = presets

-- Constants

local noise_types = {"Red (Brown)", "Pink", "White", "Blue", "Violet/Purple", "Loggy", "Base2", "ERecip"}
local generator_types = {"Noise", "Sine", "Square", "Triangle", "Sawtooth", "Pulse"}

-- State

local current_generator = 1
local autogen = renoise.Document.ObservableBoolean(false)

--- Setup Menu Stuff

renoise.tool():add_menu_entry {
  name = "--- Sample Editor:Generate Waveform",
  invoke = function()
    show_tool_window()
  end
}

--Utility

local function show_status(message)
  renoise.app():show_status(message)
  print(message)
end

--[[================================--
    STOLEN FROM @LEDGER ON forum.renoise.com
  ==================================]]--

  --------------
  --key Handler
  --------------
  local function my_keyhandler_func(dialog,key)
     --toggle lock focus hack, allows pattern ed to get key input
     --renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     --renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     return key
  end

--[[===================================--
  STOPPP STEALING YA DUMMY, okay
  =====================================]]--

-- DO the GUI
function show_tool_window()
  generate_init()
  local vb = renoise.ViewBuilder()
  local title = "MilkyWaves"
  --local buttons = {}
  local DEFAULT_DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
  local DEFAULT_CONTROL_SPACING = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
  local DEFAULT_CONTROL_HEIGHT = renoise.ViewBuilder.DEFAULT_CONTROL_HEIGHT
  local options = renoise.tool().preferences

  local content = vb:column {
    margin = DEFAULT_DIALOG_MARGIN,
    spacing = DEFAULT_CONTROL_SPACING,
    id = "mv_content",

    vb:switch {
      value = current_generator,
      --id = "mv_gen",
      width = 400,
      items = generator_types,
      tooltip = "Generator selector",
      notifier = function(ni)
        local generator = vb.views.mv_gen
        local noise_gen = vb.views.mv_noise_gen
        local shape_gen = vb.views.mv_shape_gen

        current_generator = ni
        show_status(("Generator %s is active"):format(generator_types[ni]))
        noise_gen.visible = (ni == 1)
        shape_gen.visible = (ni ~= 1)
        vb.views.anticlick_button.visible = (ni == 4 or ni == 5)
        vb.views.duty_selector.visible = (ni == 6)

        vb.views.mv_content:resize()

        if autogen.value then process_current_generator() end
      end
    },

    vb:horizontal_aligner {
      mode="center",
      spacing = DEFAULT_CONTROL_SPACING,
      id = "mv_noise_gen",
      visible = (current_generator == 1),
      vb:column {
        spacing = DEFAULT_CONTROL_SPACING,
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          vb:text {
            text = "Frequencies:"
          },
          vb:valuebox {
            bind = options.frequencies,
            min = 2.0,
            max = 500.0,
            tooltip = "Number of frequencies for noise generator",
            notifier = function() if autogen.value then process_current_generator() end end
          }
        },
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          vb:text {
            text = "Noise Type:"
          },
          vb:popup {
            width = 100,
            items = noise_types,
            bind = options.noise,
            tooltip = "Noise type for noise generator",
            notifier = function () if autogen.value then process_current_generator() end end
          }
        },
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          vb:text {
            text = "Volume:"
          },
          vb:valuebox {
            bind = options.volume,
            min = -100.0,
            max = 100.0,
            tooltip = "Volume for generator",
            notifier = function() if autogen.value then process_current_generator() end end
          }
        },
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          vb:text {
            text = "Anticlick?"
          },
          vb:checkbox {
            bind = options.anticlick,
            tooltip = "Turns on sampler anticlick when enabled",
            notifier = function()
              if autogen.value then
                process_current_generator()
              end
            end
          }
        },
      }
    },

    vb:horizontal_aligner {
      mode = "center",
      spacing = DEFAULT_CONTROL_SPACING,
      id = "mv_shape_gen",
      visible = (current_generator ~= 1),
      vb:column {
        spacing = DEFAULT_CONTROL_SPACING,
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          vb:text {
            text = "Periods:"
          },
          vb:valuebox {
            bind = options.period,
            min = 1.0,
            max = 2000.0,
            tooltip = "Number of periods for generator",
            notifier = function() if autogen.value then process_current_generator() end end
          }
        },
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          vb:text {
            text = "Volume:"
          },
          vb:valuebox {
            bind = options.volume,
            min = -100.0,
            max = 100.0,
            tooltip = "Volume for generator",
            notifier = function() if autogen.value then process_current_generator() end end
          }
        },
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          id = "duty_selector",
          vb:text {
            text = "Duty:"
          },
          vb:slider {
            bind = options.duty,
            min = 0.01,
            max = 0.99,
            tooltip = "Duty cycle for pulse generator",
            notifier = function() if autogen.value then process_current_generator() end end
          }
        },
        vb:row {
          spacing = DEFAULT_CONTROL_SPACING,
          id = "anticlick_button",
          vb:text {
            text = "Anticlick?"
          },
          vb:checkbox {
            bind = options.anticlick,
            tooltip = "Turns on sampler anticlick when enabled",
            notifier = function()
              if autogen.value then
                process_current_generator()
              end
            end
          }
        }
      }
    },

    vb:horizontal_aligner {
      --style = 'group',
      mode = 'center',
      spacing = DEFAULT_CONTROL_SPACING,
      vb:row {
        margin = DEFAULT_DIALOG_MARGIN,
        style = "group",
        vb:button {
          height = DEFAULT_CONTROL_HEIGHT * 2,
          width = DEFAULT_CONTROL_HEIGHT * 4,
          text = "Generate",
          id = "gen_button",
          --active = not autogen,

          notifier = function()
            -- Send to processing!
            process_current_generator()
          end
        },
        vb:checkbox {
          height = DEFAULT_CONTROL_HEIGHT * 2,
          width = DEFAULT_CONTROL_HEIGHT * 4,
          bind = autogen,
          notifier = function()
            if autogen.value then
              process_current_generator()
            end
          end
        }
      }
    }
  }

  -- create the basic view
  print "Called MilkyWaves.... but no one was there..."

  -- Advice and a function taken from @Ledger on forum.renoise.com
  renoise.app():show_custom_dialog(title, content, my_keyhandler_func)
end

function process_current_generator()
  local options = renoise.tool().preferences
  local rwindow = renoise.app().window
  local MIDDLE_FRAME_SAMPED = renoise.ApplicationWindow.MIDDLE_FRAME_INSTRUMENT_SAMPLE_EDITOR
  if not rwindow.instrument_editor_is_detached and rwindow.active_middle_frame ~= MIDDLE_FRAME_SAMPED then
    rwindow.active_middle_frame = MIDDLE_FRAME_SAMPED
    return
  end
  process_generator(current_generator, {
    noise = options.noise.value,
    period = options.period.value,
    volume = options.volume.value,
    frequencies = options.frequencies.value,
    anticlick = options.anticlick.value,
    duty = options.duty.value,
  })
end
